package org.processmining.plugins.cosa;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;

/**
 * @author hverbeek
 * 
 *         FLOW ::= name ["icon_id" ICONID] ["duration" DURATION] ["flow_class"
 *         FLOWCLASS] ["flow_pattern" FLOWPATTERN] ["flowext_type" FLOWEXTTYPE]
 *         ["no_duplicates" NODUPLICATES] ["no_flowext_duplicates"
 *         NOFLOWEXTDUPLICATES] ["creator" CREATOR] ["declare"] ["files" FILES]
 *         ("condition" CONDITION)* (("activity"|"call_activity") ACTIVITY)*
 *         ["declare"] ["start_act"] START ["declare"] ["end_act"] END "end"
 *         "flow"
 */
public class CosaFlow extends CosaElement {

	private String name;
	private CosaIconId iconId;
	private CosaDuration duration;
	private CosaFlowClass flowClass;
	private CosaFlowPattern flowPattern;
	private CosaFlowextType flowextType;
	private CosaNoDuplicates noDuplicates;
	private CosaNoFlowextDuplicates noFlowextDuplicates;
	private CosaCreator creator;
	private CosaFiles files;
	private final Map<String, CosaCondition> conditionMap; // Maps condition name onto condition.
	private final Map<String, CosaActivity> activityMap; // Maps activity name onto activity.
	private CosaStart start;
	private CosaEnd end;

	public CosaFlow() {
		name = null;
		iconId = null;
		duration = null;
		flowClass = null;
		flowPattern = null;
		flowextType = null;
		noDuplicates = null;
		noFlowextDuplicates = null;
		creator = null;
		files = null;
		conditionMap = new HashMap<String, CosaCondition>();
		activityMap = new HashMap<String, CosaActivity>();
		start = null;
		end = null;
	}

	/*
	 * Parse the stream for a flow. Need the path in case some activity has a
	 * subflow.
	 */
	public void parse(CosaStream input, CosaScript parent, CosaScript flow, String path) throws IOException {
		/*
		 * Get name of the flow.
		 */
		name = (parent != null ? parent.getName() + "/" : "") + input.getToken();
		input.advance();
		if (input.getToken().equals("icon_id")) {
			input.advance();
			iconId = new CosaIconId();
			iconId.parse(input);
		}
		if (input.getToken().equals("duration")) {
			input.advance();
			duration = new CosaDuration();
			duration.parse(input);
		}
		if (input.getToken().equals("flow_class")) {
			input.advance();
			flowClass = new CosaFlowClass();
			flowClass.parse(input);
		}
		if (input.getToken().equals("flow_pattern")) {
			input.advance();
			flowPattern = new CosaFlowPattern();
			flowPattern.parse(input);
		}
		if (input.getToken().equals("flowext_type")) {
			input.advance();
			flowextType = new CosaFlowextType();
			flowextType.parse(input);
		}
		if (input.getToken().equals("no_duplicates")) {
			input.advance();
			noDuplicates = new CosaNoDuplicates();
			noDuplicates.parse(input);
		}
		if (input.getToken().equals("no_flowext_duplicates")) {
			input.advance();
			noFlowextDuplicates = new CosaNoFlowextDuplicates();
			noFlowextDuplicates.parse(input);
		}
		if (input.getToken().equals("creator")) {
			input.advance();
			creator = new CosaCreator();
			creator.parse(input);
		}
		if (input.getToken().equals("declare")) {
			input.advance();
		}
		if (input.getToken().equals("files")) {
			input.advance();
			files = new CosaFiles();
			files.parse(input);
		}
		while (input.getToken().equals("condition")) {
			CosaCondition condition = new CosaCondition();
			condition.parse(input);
			conditionMap.put(condition.getName(), condition);
		}
		while (input.getToken().equals("activity") || input.getToken().equals("call_activity")) {
			CosaActivity activity = new CosaActivity();
			activity.parse(input, parent, flow, path);
			activityMap.put(activity.getName(), activity);
		}
		if (input.getToken().equals("declare")) {
			input.advance();
		}
		if (input.getToken().equals("start_act")) {
			input.advance();
			start = new CosaStart();
			start.parse(input);
		}
		if (input.getToken().equals("declare")) {
			input.advance();
		}
		if (input.getToken().equals("end_act")) {
			input.advance();
			end = new CosaEnd();
			end.parse(input);
		}
		if (input.getToken().equals("end")) {
			input.advance();
		}
		if (input.getToken().equals("flow")) {
			input.advance();
		}
	}

	public String getName() {
		return name;
	}

	/*
	 * Converts the flow into a net.
	 */
	public void convertToNet(PetrinetGraph net, ExpandableSubNet subnet, Place startPlace, Place endPlace,
			Map<String, Place> placeMap) {
		for (CosaCondition condition : conditionMap.values()) {
			condition.convertToNet(net, subnet, placeMap);
		}
		for (CosaActivity activity : activityMap.values()) {
			activity.convertToNet(net, subnet, startPlace, endPlace, activity.getName().equals(start.getName()),
					activity.getName().equals(end.getName()), placeMap);
		}
	}

}
