package org.processmining.plugins.cosa;

import java.io.IOException;
import java.io.InputStream;

import org.processmining.plugins.cosa.utils.PeekableInputStream;

/**
 * @author hverbeek
 * 
 *         Abstraction layer: converts input characters to input tokens.
 */
public class CosaStream {

	/*
	 * The last token read.
	 */
	String token;
	/*
	 * The peekable input stream. Sometimes we want to peek, not read.
	 */
	PeekableInputStream input;

	/*
	 * Create a new COSA Stream. Note that the first token is read by this.
	 */
	public CosaStream(InputStream input) throws IOException {
		this.input = new PeekableInputStream(input);
		getNextToken();
	}

	/*
	 * Read the next token from the stream.
	 */
	private void getNextToken() throws IOException {
		/*
		 * Read the fist character.
		 */
		int i = input.read();
		/*
		 * Create a fresh buffer.
		 */
		StringBuffer buffer = new StringBuffer();
		while (i >= 0) {
			/*
			 * A valid character has been read.
			 */
			/*
			 * Skip all subsequent white space characters.
			 */
			while ((i >= 0) && isSpace((char) i)) {
				i = input.read();
			}
			/*
			 * Check the first non-white-space character.
			 */
			switch ((char) i) {
				case '"' :
					/*
					 * Found a quote, that is, a String value starts here. Read
					 * until the assumed end.
					 */
					i = input.read();
					while ((i >= 0) && (((char) i) != '"')) {
						buffer.append((char) i);
						i = input.read();
					}
					/*
					 * We have now read something like "foo", with foo in the
					 * buffer. Possibly "foo" is followed by "bar", in which
					 * case we should extend the buffer to foo"bar. Check
					 * whether next character is indeed another quote.
					 */
					i = input.peek();
					while ((i >= 0) && (((char) i) == '"')) {
						/*
						 * Yes it is, extend the String.
						 */
						i = input.read();
						buffer.append((char) i);
						i = input.read();
						while ((i >= 0) && (((char) i) != '"')) {
							buffer.append((char) i);
							i = input.read();
						}
						i = input.peek();
					}
					/*
					 * No more String extensions.
					 */
					if (i >= 0) {
						/*
						 * Still okay. Return buffer as String.
						 */
						token = new String(buffer);
						return;
					}
					break;
				case '/' :
					/*
					 * Found a forward slash. Three cases: 1. First character of
					 * /*, in which case this slash starts a multi-line comment.
					 * 2. First character of //, in which case this slash starts
					 * a single-line comment. 3. None of the above, in which
					 * case it could belong to an arbitrary token.
					 */
					buffer.append((char) i);
					i = input.read();
					if (((char) i) == '*') {
						/*
						 * Found /*, read until * and /.
						 */
						boolean ast = false;
						i = input.read();
						while ((i >= 0) && (!ast || (((char) i) != '/'))) {
							ast = ((char) i) == '*';
							i = input.read();
						}
						buffer = new StringBuffer();
						/*
						 * Ignore this multi-line comment, proceed to next
						 * token.
						 */
						break;
					} else if (((char) i) == '/') {
						/*
						 * Found //
						 */
						i = input.read();
						while ((i >= 0) && (((char) i) != '\n')) {
							i = input.read();
						}
						buffer = new StringBuffer();
						/*
						 * Ignore this single-line comment, proceed to next
						 * token.
						 */
						break;
					}
					/*
					 * No comment, should start a regular token.
					 */
					//$FALL-THROUGH$
				default :
					buffer.append((char) i);
					/*
					 * Peek at the next character. If amy start a comment, don't
					 * touch it.
					 */
					i = input.peek();
					while ((i >= 0) && !isSpace((char) i) && (((char) i) != '/')) {
						input.read();
						buffer.append((char) i);
						i = input.peek();
					}
					/*
					 * Return buffer as String.
					 */
					token = new String(buffer);
					return;
			}
			i = input.read();
		}
		/*
		 * Ran out of characters to read, no token found yet. Return null.
		 */
		token = null;
	}

	/*
	 * Returns whether the given character is considered to be white space.
	 */
	private boolean isSpace(char c) {
		return Character.isWhitespace(c) || ("(),;+".indexOf(c) != -1);
	}

	/*
	 * Gets the last token read.
	 */
	public String getToken() {
		return token;
	}

	/*
	 * Advances to the next token.
	 */
	public void advance() throws IOException {
		getNextToken();
	}
}
